/*
---------------------------------------------------------------
	CRX-COM Node-js server
	Date   : 01/04/2019
	update : 22/10/2023
---------------------------------------------------------------
*/
module.exports = function(){
	
	var version			    = '1.0.4';//1.0.4-r5 (22/10/2023)
	var webSocketServerPort = 8989;
	
	var DISPLAY_DEBUG_NET   = 0;
	var DISPLAY_DEBUG_COM   = 1;
	var SILENT_MODE 		= 0;	
	
	const fs = require('fs');
	const ini = require('ini');

	const config = ini.parse(fs.readFileSync('./config.ini', 'utf-8'));	
	
	require('./crx.common')();	
		
	if(isset(config.SILENT_MODE) && !empty(config.SILENT_MODE)){SILENT_MODE=config.SILENT_MODE;}
	if(isset(config.DISPLAY_DEBUG_COM) && !empty(config.DISPLAY_DEBUG_COM)){DISPLAY_DEBUG_COM=config.DISPLAY_DEBUG_COM;}
	if(isset(config.DISPLAY_DEBUG_NET) && !empty(config.DISPLAY_DEBUG_NET)){DISPLAY_DEBUG_NET=config.DISPLAY_DEBUG_NET;}
	if(isset(config.LISTEN_PORT) && !empty(config.LISTEN_PORT)){webSocketServerPort=config.LISTEN_PORT;}
	declare(SILENT_MODE,DISPLAY_DEBUG_NET,DISPLAY_DEBUG_COM);//declare configuration for crx.common librairy. 
	
	const {SerialPort} = require('serialport');
	//const ByteLength = SerialPort.parsers.ByteLength;	not use here
	
	var final_hex 					= 	[];
	var RESPONSE_COUNTER 			=	0;
	var MAX_BUFF_LENGTH 			= 	4;
	
	//GLOBAL : 
	var currentUserConfiguration	=	{};
		
	//CAT SYSTEM : 
	var configuration_cat_is_ready	= 	0;
	var targetRunningPortConfigCat 	= 	{};
	var myserialPortCat				= 	{};
	var setPortConfigCat			=	{};
	var portNameCat					=	'';
	var connectCatId				= 	-1;
		
	//ROTATOR CONTROL:
	var configuration_rt_is_ready 		= 	0;
	var myserialPortRt	 				= 	{};
	var setPortConfigRt					=	{};
	var connectRtId						= 	-1;
	var portNameRt						=	'';
	
	//CW KEYER : 
	var configuration_cw_is_ready 		= 	0;
	var myserialPortCw	 				= 	{};	
	var setPortConfigCw					=	{};
	var connectCwId						= 	-1;
	var portNameCw						=	'';	
	
	//HTTPS Layer : 
	var express = require('express');     
	var sleep = require('system-sleep');
	var privateKey = fs.readFileSync('sslcert/key.pem');
	var certificate = fs.readFileSync('sslcert/cert.pem');
	var https = require('https');
	var WebSocket = require('ws');        
	var WebSocketServer = WebSocket.Server;
	var server = express();                
	var credentials = {key: privateKey, cert: certificate};
	var httpServer = https.createServer(credentials,server); // an http server

	var wss = new WebSocketServer({ server: httpServer }); // a websocket server
	
	wss.on("error",function(evt){
		console.log('Error Occured, if EADDRINUSE, change the port inside config.ini, and in your online CAT configuration.');
		console.log(evt);
	 }
	);
	
	server.use('/',express.static('www'));//if needed

	//60wpm => 20   ms 
	//10wpm => 120  ms
	var cw_element_lenght_default 	= 	40;	
						
	function setCwPortStatus(port,typePort,statusPort,debugWithoutCom){    
		
		if(typePort=='RTS'){var type='rts';}
		if(typePort=='DTR'){var type='dtr';}	
		if(debugWithoutCom==1){
			return new Promise(
				function(resolve, reject) {       
					console.log('debug without com set:'+type+':'+statusPort)        
					resolve("done!");					
				}
			);			
		}
		else
			return new Promise(
				function(resolve, reject) {       
					port.set({type:statusPort}, function () {			
						console.log('set:'+type+':'+statusPort)        
						resolve("done!");					
					});
				}
			);
	}

	function connectClient(newClient) {	   		
		
		debugc('@wait for client websocket ...',DISPLAY_DEBUG_NET);
		
		async function readMessage(data) {			
						
			var obj = JSON.parse(data);

			if('getRxMeterData' in obj){
				
				var cmd			=	'getRxMeterData';
				var protocol	=	obj.protocol;	
				
				final_hex = [];
				MAX_BUFF_LENGTH = obj.readbuff;
				RESPONSE_COUNTER=0;	
				
				debugc('@debug:'+obj.protocol,DISPLAY_DEBUG_COM);			
				debugc('@getRxMeterData,info,input data:'+obj.getRxMeterData,DISPLAY_DEBUG_COM);				
				debugc('@getRxMeterData,readbuff:'+obj.readbuff,DISPLAY_DEBUG_COM);
				debugc('@getRxMeterData,MAX_BUFF_LENGTH:'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);
				
				if(!myserialPortCat instanceof SerialPort){
					return newClient.send('ko,not connected to serial port.',function(error){if(error != null){}});
				}
	
				var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.getRxMeterData)); 				
								
				var rsp = {};
				
				rsp.smeter = '00';
				
				if(connectCatId){					
					if(isFunction(myserialPortCat.removeAllListeners)){
						debugc('@debug:serialPort.removeAllListeners',DISPLAY_DEBUG_COM);					
						myserialPortCat.removeAllListeners("data");
					}
					else{
						debugc('@debug:serialPort broken connection.',DISPLAY_DEBUG_COM);						
						return;
					}
				}
				
				//decode incoming data : 
				if(isFunction(myserialPortCat.on)){				
					
					//=>   icom1, yaesu1, yaesu3
					if(protocol=='yaesu3' || protocol=='yaesu1' || protocol=='icom1'){						
						myserialPortCat.on('data', function(chunk){																
							debugc('@debug,incoming data ...',DISPLAY_DEBUG_COM);												
							if(cmd != 'getRxMeterData' || MAX_BUFF_LENGTH == 0){
								debugc('@debug,incoming data => return.',DISPLAY_DEBUG_COM);
								return;
							}
							var arr = new Uint8Array(chunk);
							for (byte in arr) {
								var b=arr[byte].toString(16);
								var bout='';
								if(b.length==1)bout='0'+b;
								else bout+=b;
								final_hex.push(bout);							
							}							
							RESPONSE_COUNTER = RESPONSE_COUNTER + 1;											
							debugc('RX ====== HEX ====== RS:'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);						
							debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);					
							debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);					
							debugc(final_hex,DISPLAY_DEBUG_COM);							
							
							if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){														
								debugc('====== END ======',DISPLAY_DEBUG_COM);						
								debugc('@getRxMeterData,protocol,'+protocol,DISPLAY_DEBUG_COM);															
								//@TODO 
								/*if(protocol=='yaesu3'){		
									...
								}*/
								if(protocol=='yaesu1'){							
									//ex: [ '09' ]  ,      [ '2e' ] 
									//console.log(final_hex);
									var report = '';
									for (v in final_hex) {
										report+=final_hex[v];
									}//ex 2e => 46 
									report = String(parseInt(report,16));
									rsp.smeter=report;
									debugc('@getRxMeterData,'+protocol+',reply to client report:'+report,DISPLAY_DEBUG_COM);	
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){c('getRxMeterData,websocket,error: %s', error);}
									})
								}														
								if(protocol=='icom1'){
									// Reply contains one or two byte of data, 0=0%, 255=100% on S-Meter
									var report = '';
									report = icom1_getRxMeterData(final_hex);																				       rsp.smeter=report;							
									debugc('@getRxMeterData,'+protocol+',reply to client:'+report,DISPLAY_DEBUG_COM);															       newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('getRxMeterData,websocket,error:'+error,DISPLAY_DEBUG_COM);									
										}
									});													
									final_hex = [];
									RESPONSE_COUNTER=0;										
								}
								return;
							}						
					      }
						);					
					}					
					//kenwood1, elecraft1, yaesu2 : 
					else{						
						// @NOT-TESTED:
						debugc('@debug,define myserialPortCat.on(data) ...',DISPLAY_DEBUG_COM);
						myserialPortCat.on('data', function(chunk){													
							
							debugc('@debug,incoming data ...',DISPLAY_DEBUG_COM);
							
							if(cmd != 'getRxMeterData' || MAX_BUFF_LENGTH == 0)return;						
							
							var arr = new Uint8Array(chunk);
							
							for (byte in arr) {
								var b=arr[byte].toString(16);
								var bout='';
								if(b.length==1)bout='0'+b;
								else bout+=b;
								final_hex.push(bout);							
							}

							RESPONSE_COUNTER = RESPONSE_COUNTER + 1;											

							debugc('RX ====== HEX ====== RS'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);						
							debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);					
							debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);					
							debugc(final_hex,DISPLAY_DEBUG_COM);							

							if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){
								debugc('====== END ======',DISPLAY_DEBUG_COM);						
								debugc('@getRxMeterData,protocol,'+protocol,DISPLAY_DEBUG_COM);									
								debugc('@getRxMeterData,value:,'+chunk,DISPLAY_DEBUG_COM);

								if(protocol=='yaesu2'){	
									debugc('@getRxMeterData, extract : '+protocol+', final_hex value:'+final_hex,DISPLAY_DEBUG_COM);	  	 							
									/* @TODO
									var qrg = yaesu2_get_qrg(final_hex);																
									qrg = parseInt(qrg,10);									
									rsp.frq=qrg;																
									debugc('@getRxMeterData,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);	  	 							
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('@getRxMeterData,websocket,error:'+error,DISPLAY_DEBUG_COM);
										}
									});	
									myserialPortCat.removeAllListeners("data");	
									*/
								}	
								
								if(protocol=='kenwood1'|| protocol=='elecraft1'){
									/* @TODO
									debugc('@getRxMeterData, extract : '+protocol+', final_hex value:'+final_hex,DISPLAY_DEBUG_COM);	  	 							
									
									//ex:	FA00007000000; not tested 																								
									var qrg = kenwood1_get_qrg(final_hex);																
									qrg = parseInt(qrg,10);									
									rsp.frq=qrg;								
									debugc('@getRxMeterData,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);								
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('@getRxMeterData,websocket,error:'+error,DISPLAY_DEBUG_COM);
										}
									});
									myserialPortCat.removeAllListeners("data");									
									final_hex = [];
									RESPONSE_COUNTER=0;	*/
								}							
							
							}
						
						
						});						
					
					}					
				}

				//write string  @NOT-TESTED:
				if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
					debugc('@getRxMeterData,info,write method is : string ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(obj.getRxMeterData,function(err){if(err){return debugc('@getRxMeterData,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}else{debugc('@getRxMeterData,info,serialPort.write: '+obj.getRxMeterData,DISPLAY_DEBUG_COM);}});										
				}
				//write buffer ex yaesu1 : 
				else{
					debugc('@getRxMeterData,info,write method is : buffer ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(Buffer.from(bytes_to_send.buffer),function(err){
						if (err) {
							return debugc('ERROR, getRxMeterData,error,serialPort.write:'+err.message,DISPLAY_DEBUG_COM);
						}
						else{
							debugc('OK, getRxMeterData,info,serialPort.write:'+Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);	
						}
					});
				}
			}
			
			if('setRotator' in obj){

				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@setRotator,get new rotor position cmd:'+obj.setRotator,DISPLAY_DEBUG_COM);
				debugc('@setRotator,info,protocol:'+currentUserConfiguration['rotorProtocol'],DISPLAY_DEBUG_COM);	
								
				if(((myserialPortRt instanceof SerialPort) === false  ) || ( connectRtId == -1)){
					var m='@setRotator,error,not connected to serial port.';
					newClient.send(m);
					debugc(m,DISPLAY_DEBUG_COM);	
					return;
				}
				else{
					if(	connectRtId	!=	-1	){
						var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.setMode)); 		
						myserialPortRt.removeAllListeners("data");
						if(currentUserConfiguration['rotorProtocol']=='GS23'){
							var cmdSetAngle   = 'M';
							var angle = sprintf('%03d',obj.setRotator);
							myserialPortRt.write(cmdSetAngle+angle+"\r\n");
							return;
						}
						if(currentUserConfiguration['rotorProtocol']=='EZ'){
							var cmdSetAngle   = 'AP1';
							var angle = sprintf('%03d',obj.setRotator);
							myserialPortRt.write(cmdSetAngle+angle+"\r\n");
							return;					
						}
						if(currentUserConfiguration['rotorProtocol']=='DCU'){//based on rotorez.c hamlib: 
							if (obj.setRotator > 359.4999)		/* DCU-1 compatibility */
								obj.setRotator = 0;
							var cmdSetAngle   = 'AP1';//Target bearing
							var angle = sprintf('%03d',obj.setRotator);
							myserialPortRt.write(cmdSetAngle+angle+";\r\n");					
							var cmdExec   = 'AM1;';//Execute command
							myserialPortRt.write(cmdExec+"\r\n");					
							return;	
						}
					}
				}
			}
			
			if('cwkeym' in obj){
					
				//@WORK IN PROGRESS  : 
	
				//This can be used to disabled serial port communication, 
				//simply display morse char on console. 				
				this.debugWithoutCom = 1; 
				
				if(this.debugWithoutCom==1){
					connectCwId=1;
				}
				
				if(  this.debugWithoutCom == 0 ||   ( (		(myserialPortCw instanceof SerialPort) === false  ) || ( connectCwId == -1  ) ) ){
					var m='@cwkeym,error,not connected to serial port.';
					newClient.send(m);
					c(m);	
					return;
				}			
				else if(connectCwId != -1){
					debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
					debugc('@cwkeym,get new CW message to send.',DISPLAY_DEBUG_COM);
					debugc('- currentUserConfiguration cwWpmSpeed:'+currentUserConfiguration['cwWpmSpeed'],DISPLAY_DEBUG_COM);
					debugc('- currentUserConfiguration comPortCwType:'+currentUserConfiguration['comPortCwType'],DISPLAY_DEBUG_COM);	
					//myserialPortCw	=>  the serial port instance. 
					var time_unit 			= cw_element_lenght_default;
					if(isset(currentUserConfiguration['cwWpmSpeed'])){
						//Formular :        1200 / [WPM]  =  [ELEMENT-LENGTH-MS] 	
						var wpm = parseInt(currentUserConfiguration['cwWpmSpeed'],10);
						time_unit = Math.round(1200 / wpm);	
					}
					debugc('=> calcul element lenght in ms:'+time_unit,DISPLAY_DEBUG_COM);	
					var time_dot			= time_unit 	* 	1;//Length of dot
					var time_dash 			= time_unit 	* 	3;//dash
					var time_between_char 	= time_unit 	* 	3;//between characters
					var time_between_words 	= time_unit 	* 	7;//between words
					var time_delay = 0;	
					var str=obj.cwinmess;
					////////////////
					//Serial =>  open port ex:  /dev/ttyUSB0
					//	=>   set RTS => 0
					////////////////				
					//ex data :   -.-.--.-    /       -.-.--.-       /   ..-.....-.-.----.-
					for (var i = 0; i < str.length; i++) {
					  time_delay = 0;
					  var dotdash = str.charAt(i);
					  if (dotdash == '.') { // => RX:  .   => delay 1 
						time_delay = time_dot;
						//set RTS => 1 
						//if(currentUserConfiguration['comPortCwType']=='RTS' && this.debugWithoutCom==0)myserialPortCw.set({rts:true});
						//if(currentUserConfiguration['comPortCwType']=='DTR' && this.debugWithoutCom==0)myserialPortCw.set({dtr:true});						
						
						var setStatus = await setCwPortStatus(myserialPortCw,currentUserConfiguration['comPortCwType'],true,this.debugWithoutCom);						
						sleep(time_delay);
						
						//set RTS => 0 
						//if(currentUserConfiguration['comPortCwType']=='RTS' && this.debugWithoutCom==0)myserialPortCw.set({rts:false});
						//if(currentUserConfiguration['comPortCwType']=='DTR' && this.debugWithoutCom==0)myserialPortCw.set({dtr:false});
						
						var setStatus = await setCwPortStatus(myserialPortCw,currentUserConfiguration['comPortCwType'],false,this.debugWithoutCom);						
						debugc('->tx : '+dotdash+"\r",DISPLAY_DEBUG_COM);
						sleep(time_between_char);
						
					  }
					  else if (dotdash == '-') { // => RX:  - => delay 3 
						time_delay = time_dash;
						//set RTS => 1
						//if(currentUserConfiguration['comPortCwType']=='RTS' && this.debugWithoutCom==0)myserialPortCw.set({rts:true});
						//if(currentUserConfiguration['comPortCwType']=='DTR' && this.debugWithoutCom==0)myserialPortCw.set({dtr:true});
												
						var setStatus = await setCwPortStatus(myserialPortCw,currentUserConfiguration['comPortCwType'],true,this.debugWithoutCom);						
						sleep(time_delay);						
						
						//set RTS => 0
						//if(currentUserConfiguration['comPortCwType']=='RTS' && this.debugWithoutCom==0)myserialPortCw.set({rts:false});
						//if(currentUserConfiguration['comPortCwType']=='DTR' && this.debugWithoutCom==0)myserialPortCw.set({dtr:false});						
						
						var setStatus = await setCwPortStatus(myserialPortCw,currentUserConfiguration['comPortCwType'],false,this.debugWithoutCom);						
						debugc('->tx : '+dotdash+"\r",DISPLAY_DEBUG_COM);//console display. 			
						sleep(time_between_char);
						
						
					  }
					  else { // => RX:  /  => delay 7
						time_delay = time_between_words;
						debugc('wait ... '+"\r",DISPLAY_DEBUG_COM);
						sleep(time_delay);					
					  }
					}				
					newClient.send('ok',function(error){if(error != null){}})				
				}				
			}

			//using also for PTT cmd
			if('setTxOff' in obj){
				
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@setTxOff',DISPLAY_DEBUG_COM);
				
				if(!myserialPortCat instanceof SerialPort){
					return newClient.send('ko,not connected to serial port.',function(error){if(error != null){}});
				}	
				
				if(connectCatId){
					debugc('@setTxOff,info : removeAllListeners',DISPLAY_DEBUG_COM);					
					myserialPortCat.removeAllListeners("data");
				}
				
				myserialPortCat.on('data', function(chunk){});
				
				var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.setTxOff));
				
				//write string 
				if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
					debugc('@setTxOff,info,write method is : string ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(obj.getMode,function(err){if(err){return debugc('@setTxOff,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}else{debugc('@setTxOff,info,serialPort.write: '+obj.setTxOff,DISPLAY_DEBUG_COM);}});					
				}
				//write buffer 
				//if(protocol=='icom1' || protocol=='yaesu1'){
				else{	
					//ex             Buffer.from([0xA9, 0x17, 0x2F, 0x00, 0x ... 
					debugc('@setTxOff,info,write method is : buffer ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(Buffer.from(bytes_to_send.buffer),function(err){if (err){return debugc('@setTxOff,error,serialPort.write: ', err.message,DISPLAY_DEBUG_COM);}else{debugc('@setTxOff,info,serialPort.write: '+Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);}});	 
				}				
			}

			//using also for PTT cmd
			if('setTxOn' in obj){
				
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@setTxOn',DISPLAY_DEBUG_COM);
				
				if(!myserialPortCat instanceof SerialPort){
					return newClient.send('ko,not connected to serial port.',function(error){if(error != null){}});
				}	
				
				if(connectCatId){
					debugc('@setTxOn,info : removeAllListeners',DISPLAY_DEBUG_COM);					
					myserialPortCat.removeAllListeners("data");
				}
				
				myserialPortCat.on('data', function(chunk){});
				
				var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.setTxOn));
				//write string 
				if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
					debugc('@setTxOn,info,write method is : string ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(obj.getMode,function(err){if(err){return debugc('@setTxOn,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}else{debugc('@setTxOn,info,serialPort.write: '+obj.getMode,DISPLAY_DEBUG_COM);}});					
				}
				//write buffer 
				//=>   icom1, yaesu1, yaesu3
				else{	
					//ex             Buffer.from([0xA9, 0x17, 0x2F, 0x00, 0x ... 
					debugc('@setTxOn,info,write method is : buffer ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(Buffer.from(bytes_to_send.buffer),function(err){if (err){return debugc('@setTxOn,error,serialPort.write: ', err.message,DISPLAY_DEBUG_COM);}else{debugc('@setTxOn,info,serialPort.write: '+Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);}});	 
				}				
			}

			if('getMode' in obj){				
								
				var cmd			=	'getMode';								
				var protocol	=	obj.protocol;					
				final_hex 		= 	[];				
				MAX_BUFF_LENGTH = obj.readbuff;				
				RESPONSE_COUNTER=0;					
				
				c('@debug:'+obj.protocol);			
				c('@getMode,info,input data:'+obj.getMode);				
				c('@getMode,readbuff:'+obj.readbuff);				
				if(!myserialPortCat instanceof SerialPort){
					return newClient.send('ko,not connected to serial port.',function(error){if(error != null){}});
				}							
				var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.getMode)); 				
				var rsp = {};
				rsp.mode = '';				
				if(connectCatId){
					debugc('@serialPort,info : removeAllListeners',DISPLAY_DEBUG_COM);					
					myserialPortCat.removeAllListeners("data");
				}
				
				//=>   icom1, yaesu1, yaesu3
				if(protocol=='yaesu3' || protocol=='yaesu1' || protocol=='icom1'){
					myserialPortCat.on('data', function(chunk){									
							debugc('@serialPort,info : incoming data ...',DISPLAY_DEBUG_COM);							
							if(cmd != 'getMode' || MAX_BUFF_LENGTH == 0)return;
							var arr = new Uint8Array(chunk);
							for (byte in arr) {
								var b=arr[byte].toString(16);
								var bout='';
								if(b.length==1)bout='0'+b;
								else bout+=b;
								final_hex.push(bout);							
							}						
							debugc('RX ====== HEX ====== RS'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);	
							RESPONSE_COUNTER = RESPONSE_COUNTER + 1;										
							debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);	
							debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);							
							
							if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){													
								debugc('====== END ======',DISPLAY_DEBUG_COM);
								debugc('@getMode,protocol,'+protocol,DISPLAY_DEBUG_COM);															
								
								if(protocol=='yaesu3'){return;} // with yaesu 1/3 mode is return on get frequency ( all data are in same response ). 
								if(protocol=='yaesu1'){return;}																				
								
								if(protocol=='icom1'){																						
									var mode = icom1_get_mode(final_hex);																					
									rsp.mode	=	mode;							
									debugc('@getMode,'+protocol+',reply to client:'+mode,DISPLAY_DEBUG_COM);									
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('getMode,websocket,error:'+error,DISPLAY_DEBUG_COM);									
										}
									});													
									final_hex = [];
									RESPONSE_COUNTER=0;																						
								}														
								return;								
							}						
					   }
					);				
				}
				
				else{					
					//kenwood1, elecraft1, yaesu2 : 	
					myserialPortCat.on('data', function(chunk){									
						if(cmd != 'getMode' || MAX_BUFF_LENGTH == 0)return;						
						var arr = new Uint8Array(chunk);
						for (byte in arr) {
							var b=arr[byte].toString(16);
							var bout='';
							if(b.length==1)bout='0'+b;
							else bout+=b;
							final_hex.push(bout);							
						}
							
						debugc('@debug,incoming data ...',DISPLAY_DEBUG_COM);						
						debugc('@getMode,protocol,'+protocol,DISPLAY_DEBUG_COM);
						debugc('@getMode,value:,'+chunk,DISPLAY_DEBUG_COM);

						RESPONSE_COUNTER = RESPONSE_COUNTER + 1;											
						debugc('RX ====== HEX ====== RS'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);						
						debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);					
						debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);					
						debugc(final_hex,DISPLAY_DEBUG_COM);							

						if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){
							debugc('====== END ======',DISPLAY_DEBUG_COM);						
							debugc('@getMode,protocol,'+protocol,DISPLAY_DEBUG_COM);															

							debugc('@serialPort,info : incoming data ...',DISPLAY_DEBUG_COM);							
							if(cmd != 'getMode')return;								
							if(protocol=='yaesu2'){
								var mode = yaesu2_get_mode(final_hex);																					
								rsp.mode	=	mode;							
								debugc('@getMode,'+protocol+',reply to client:'+mode,DISPLAY_DEBUG_COM);									
								newClient.send(JSON.stringify(rsp),function(error){
									if(error != null){
										debugc('getMode,websocket,error:'+error,DISPLAY_DEBUG_COM);									
									}
								});													
							}							
							//not tested: 
							if(protocol=='kenwood1' || protocol=='elecraft1'){
								var mode = kenwood1_get_mode(final_hex);																					
								rsp.mode	=	mode;							
								debugc('@getMode,'+protocol+',reply to client:'+mode,DISPLAY_DEBUG_COM);									
								newClient.send(JSON.stringify(rsp),function(error){
									if(error != null){
										debugc('getMode,websocket,error:'+error,DISPLAY_DEBUG_COM);									
									}
								});													
							}						
						}					
					});
					
				}
				
				//write string 
				if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
					debugc('@getMode,info,write method is : string ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(obj.getMode,function(err){if(err){return debugc('@getMode,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}else{debugc('@getMode,info,serialPort.write: '+obj.getMode,DISPLAY_DEBUG_COM);}});					
				}
				//write buffer 
				//=>   icom1, yaesu1, yaesu3
				else{	
					//ex             Buffer.from([0xA9, 0x17, 0x2F, 0x00, 0x ... 
					debugc('@getMode,info,write method is : buffer ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(Buffer.from(bytes_to_send.buffer),function(err){if (err){return debugc('@getMode,error,serialPort.write: ', err.message,DISPLAY_DEBUG_COM);}else{debugc('@getMode,info,serialPort.write: '+Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);}});	 
				}

			}
			
			if('getQrg' in obj){
				
				var cmd			=	'getQrg';
				var protocol	=	obj.protocol;	
				
				final_hex = [];
				MAX_BUFF_LENGTH = obj.readbuff;
				RESPONSE_COUNTER=0;	
				
				debugc('@debug:'+obj.protocol,DISPLAY_DEBUG_COM);			
				debugc('@getQrg,info,input data:'+obj.getQrg,DISPLAY_DEBUG_COM);				
				debugc('@getQrg,readbuff:'+obj.readbuff,DISPLAY_DEBUG_COM);
				debugc('@getQrg,MAX_BUFF_LENGTH:'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);
				
				if(!myserialPortCat instanceof SerialPort){
					return newClient.send('ko,not connected to serial port.',function(error){if(error != null){}});
				}
	
				var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.getQrg)); 				
				var rsp = {};rsp.frq = '0000';
				
				if(connectCatId){					
					if(isFunction(myserialPortCat.removeAllListeners)){
						debugc('@debug:serialPort.removeAllListeners',DISPLAY_DEBUG_COM);					
						myserialPortCat.removeAllListeners("data");
					}
					else{
						debugc('@debug:serialPort broken connection.',DISPLAY_DEBUG_COM);						
						return;
					}
				}
				
				//decode incoming data : 
				if(isFunction(myserialPortCat.on)){				
					
					//=>   icom1, yaesu1, yaesu3
					if(protocol=='yaesu3' || protocol=='yaesu1' || protocol=='icom1'){						
						
						myserialPortCat.on('data', function(chunk){																
							debugc('@debug,incoming data ...',DISPLAY_DEBUG_COM);												
							if(cmd != 'getQrg' || MAX_BUFF_LENGTH == 0){
								debugc('@debug,incoming data => return.',DISPLAY_DEBUG_COM);
								return;
							}
							var arr = new Uint8Array(chunk);
							for (byte in arr) {
								var b=arr[byte].toString(16);
								var bout='';
								if(b.length==1)bout='0'+b;
								else bout+=b;
								final_hex.push(bout);							
							}							
							RESPONSE_COUNTER = RESPONSE_COUNTER + 1;											
							debugc('RX ====== HEX ====== RS'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);						
							debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);					
							debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);					
							debugc(final_hex,DISPLAY_DEBUG_COM);							
							
							if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){														
								debugc('====== END ======',DISPLAY_DEBUG_COM);						
								debugc('@getQrg,protocol,'+protocol,DISPLAY_DEBUG_COM);															
								
								//@BETA : 
								if(protocol=='yaesu3'){		
									var qrg='';	
									var modeid='';
									var j=0;	
									for (v in final_hex) {													
										
										//   0      1 - 4    5-6     7    9 - F
										//  BAND  FREQUENCY  CLAR.  MODE  xxxxx
									    // 
										
										//ex ft8x7 serie : [ '00', '71', '91', '50', '0a' ]
										//ex ft1000 serie : xxxxxxxxxxxxxxxxxxx @COMPLETE 
										//  [        qrg          ][ mode ]								
										
										if(j>=1 && j<=4)qrg+=final_hex[v];
										if(j==4)modeid=final_hex[v];
										
										//if(j==4)modeid=final_hex[v]; 
										//remove last bloc(4) 0a. 
										//if(j<=3)qrg+=final_hex[v];
										j++;//   
									}
									//@TODO : QRG Conversion 
									//ex: //    7191    719150  // * 10   ??? 
									//////////////////////////////////
									//Operating frequency data : 
									//  B1  B2  B3  B4 
									//
									//  00  50  42  01  H  =>  1,425,000 (base conversion)
									// 
									// 1,425,000 Hz   =>   14.250.00 Mhz 
									//
									//////////////////////////////////

									/*
										var buff = final_hex.toString('hex');		
										var out = buff.split(',');			
										var qrg = '';
										console.log(out.length);
										console.log('out:'	+	out);
										for (var i = 0; i < out.length; i++) {		
											if(out[i]!=46 && out[i]!=41 && out[i]!=42 && out[i]!='3b'){
												try{
													//console.log('d:'+decodeURI('%'+out[i]));
													qrg+=decodeURI('%'+out[i]);
												}
												catch(err) {
													//console.log(err.message);
												}			
											}		
										}		
										console.log('qrg:'+qrg);
										var qrg = parseInt(qrg,10);	
										qrg/=10;
										console.log('kenwood1_get_qrg:'+qrg);			
										return qrg;										
									*/

									
									rsp.frq=qrg;
									rsp.mode=modeid;
									debugc('@getQrg,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);						
									debugc('@getQrg,'+protocol+',reply to client mode:'+modeid,DISPLAY_DEBUG_COM);						
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){c('getQrg,websocket,error: %s', error);}
									});														
									final_hex = [];
									RESPONSE_COUNTER=0;	
									
									
									
								}
								
								if(protocol=='yaesu1'){							
									var qrg='';	
									var modeid='';
									var j=0;	
									for (v in final_hex) {													
										//[ '00', '71', '91', '50', '0a' ]
										//  [        qrg          ][ mode ]								
										if(j==4)modeid=final_hex[v]; 
										//remove last bloc(4) 0a. 
										if(j<=3)qrg+=final_hex[v];
										j++;//   
									}
									//    7191    719150
									rsp.frq=qrg;
									rsp.mode=modeid;
									debugc('@getQrg,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);						
									debugc('@getQrg,'+protocol+',reply to client mode:'+modeid,DISPLAY_DEBUG_COM);						
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){c('getQrg,websocket,error: %s', error);}
									});														
									final_hex = [];
									RESPONSE_COUNTER=0;		
								}														
								if(protocol=='icom1'){
									/*
									430mhz:
									[ 'fe', 'fe', '58',
									  'e0',  '03', 'fd',
									  'fe',  'fe', 'e0', '58',
									  '03',  '00', '00', '00',
									  '30',  '04', 'fd' ]
									*/																
									var qrg = icom1_get_qrg(final_hex);								
									qrg = parseInt(qrg,10);	
											
									//console.log('qrg:'+qrg);
									//console.log('final_hex:'+final_hex);							
									rsp.frq	=	qrg	* 10;
									
									debugc('getQrg,'+protocol+',reply to client:'+qrg,DISPLAY_DEBUG_COM);						
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('getQrg,websocket,error:'+error,DISPLAY_DEBUG_COM);
										}
									});													
									final_hex = [];
									RESPONSE_COUNTER=0;																						
								}						
								return;
							}						
					      }
						);					
					
					}					
					//kenwood1, elecraft1, yaesu2 : 
					else{						
						
						debugc('@debug,define myserialPortCat.on(data) ...',DISPLAY_DEBUG_COM);
						myserialPortCat.on('data', function(chunk){													
							
							debugc('@debug,incoming data ...',DISPLAY_DEBUG_COM);
							
							if(cmd != 'getQrg' || MAX_BUFF_LENGTH == 0)return;						
							
							var arr = new Uint8Array(chunk);
							
							for (byte in arr) {
								var b=arr[byte].toString(16);
								var bout='';
								if(b.length==1)bout='0'+b;
								else bout+=b;
								final_hex.push(bout);							
							}

							RESPONSE_COUNTER = RESPONSE_COUNTER + 1;											

							debugc('RX ====== HEX ====== RS'+RESPONSE_COUNTER,DISPLAY_DEBUG_COM);						
							debugc('=>final_hex.length'+final_hex.length,DISPLAY_DEBUG_COM);					
							debugc('=>MAX_BUFF_LENGTH'+MAX_BUFF_LENGTH,DISPLAY_DEBUG_COM);					
							debugc(final_hex,DISPLAY_DEBUG_COM);							

							if((RESPONSE_COUNTER == MAX_BUFF_LENGTH) || final_hex.length >= MAX_BUFF_LENGTH){
								debugc('====== END ======',DISPLAY_DEBUG_COM);						
								debugc('@getQrg,protocol,'+protocol,DISPLAY_DEBUG_COM);									
								debugc('@getQrg,value:,'+chunk,DISPLAY_DEBUG_COM);

								if(protocol=='yaesu2'){	
									
									//@getQrg, extract : yaesu2, final_hex value:46,41,30,37,31,35,35,30,30,30,3b
									//buff:46,41,30,37,31,35,35,30,30,30,3b
									debugc('@getQrg, extract : '+protocol+', final_hex value:'+final_hex,DISPLAY_DEBUG_COM);	  	 							
									
									//tested     FA xxxxxx;								
									var qrg = yaesu2_get_qrg(final_hex);																
									qrg = parseInt(qrg,10);									
									rsp.frq=qrg;																
									debugc('@getQrg,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);	  	 							
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('@getQrg,websocket,error:'+error,DISPLAY_DEBUG_COM);
										}
									});	
									myserialPortCat.removeAllListeners("data");	
								}	
								
								if(protocol=='kenwood1'|| protocol=='elecraft1'){
									
									debugc('@getQrg, extract : '+protocol+', final_hex value:'+final_hex,DISPLAY_DEBUG_COM);	  	 							
									
									//ex:	FA00007000000; not tested 																								
									var qrg = kenwood1_get_qrg(final_hex);																
									qrg = parseInt(qrg,10);									
									rsp.frq=qrg;								
									debugc('@getQrg,'+protocol+',reply to client frq:'+qrg,DISPLAY_DEBUG_COM);								
									newClient.send(JSON.stringify(rsp),function(error){
										if(error != null){
											debugc('@getQrg,websocket,error:'+error,DISPLAY_DEBUG_COM);
										}
									});
									myserialPortCat.removeAllListeners("data");									
									final_hex = [];
									RESPONSE_COUNTER=0;									
								}							
							
							}
						
						
						});							
					
					}					
				}

				//write string 
				if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
					debugc('@getQrg,info,write method is : string ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(obj.getQrg,function(err){
						if (err) {return debugc('@getQrg,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}
						else{debugc('@getQrg,info,serialPort.write: '+obj.getQrg,DISPLAY_DEBUG_COM);}					
					});					
				}
				//write buffer 
				else{
					debugc('@getQrg,info,write method is : buffer ',DISPLAY_DEBUG_COM);
					myserialPortCat.write(Buffer.from(bytes_to_send.buffer),function(err){
						if (err) {
							return debugc('getQrg,error,serialPort.write:'+err.message,DISPLAY_DEBUG_COM);
						}
						else{
							debugc('getQrg,info,serialPort.write:'+Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);	
						}
					});
				}
			}

			if("setMode" in obj){
				
				var cmd			=	'setMode';			
				var protocol	=	obj.protocol;
				
				final_hex = [];
				MAX_BUFF_LENGTH = obj.readbuff;
				RESPONSE_COUNTER=0;				
				
				debugc('@setMode',DISPLAY_DEBUG_COM);
				debugc('@setMode,readbuff:'+obj.readbuff,DISPLAY_DEBUG_COM);			
				
				if(  (		(myserialPortCat instanceof SerialPort) === false  ) || ( connectCatId == -1  ) ){
					var m='@setMode,error,not connected to serial port.';
					newClient.send(m);
					debugc(m,DISPLAY_DEBUG_COM);	
					return;
				}			
				else{
					if(	connectCatId	!=	-1	){
						debugc('@setMode,info,input data:'+obj.setMode,DISPLAY_DEBUG_COM);	
						var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.setMode)); 		
						myserialPortCat.removeAllListeners("data");					
						
						//write string 
						if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
							debugc('@setMode,info,write method is : string ',DISPLAY_DEBUG_COM);
							myserialPortCat.write(obj.setMode,function(err){
								if (err) {return debugc('@setMode,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}
								else{debugc('@setMode,info,serialPort.write: '+obj.setMode,DISPLAY_DEBUG_COM);}					
							});					
						}
						//write buffer 
						else{		
							myserialPortCat.write(Buffer.from(bytes_to_send.buffer));
							debugc("Transmitted data =", Buffer.from(bytes_to_send.buffer),DISPLAY_DEBUG_COM);
							newClient.send('ok');
						}
						
					}
					else{
						debugc('@setMode,error,serial port is disconnected',DISPLAY_DEBUG_COM);	
					}
				}
			}

			if("setQrg" in obj){
				var cmd			=	'setQrg';			
				var protocol	=	obj.protocol;
				
				final_hex = [];
				MAX_BUFF_LENGTH = obj.readbuff;
				RESPONSE_COUNTER=0;				
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@setQrg',DISPLAY_DEBUG_COM);
				debugc('@setQrg,readbuff:'+obj.readbuff,DISPLAY_DEBUG_COM);
				
				if(  (		(myserialPortCat instanceof SerialPort) === false  ) || ( connectCatId == -1  ) ){
					var m='@setQrg,error,not connected to serial port.';
					newClient.send(m);
					c(m);	
					return;
				}			
				else{
					if(	connectCatId	!=	-1	){
						debugc('@setQrg,info,input data:'+obj.setQrg,DISPLAY_DEBUG_COM);	
						var bytes_to_send = new Uint8Array(convertHexStrToArrayHex(obj.setQrg)); 		
						myserialPortCat.removeAllListeners("data");					
						
						//write string 
						if(protocol=='yaesu2' || protocol=='kenwood1' || protocol=='elecraft1'){
							debugc('@setQrg,info,write method is : string ',DISPLAY_DEBUG_COM);
							myserialPortCat.write(obj.setQrg,function(err){
								if (err) {return debugc('@setQrg,error,serialPort.write: ',err.message,DISPLAY_DEBUG_COM);}
								else{debugc('@setQrg,info,serialPort.write: '+obj.setQrg,DISPLAY_DEBUG_COM);}					
							});					
						}
						//write buffer 
						else{
							debugc('@setQrg,info,write method is : buffer ',DISPLAY_DEBUG_COM);
							myserialPortCat.write(Buffer.from(bytes_to_send.buffer));
							newClient.send('ok');
						}
						
					}
					else{
						debugc('@setQrg,error,serial port is disconnected',DISPLAY_DEBUG_COM);	
					}
				}
			}

			if("setConf" in obj) {				
				
				currentUserConfiguration = obj.setConf;				
				
				/*
example, rx client :  
rx client data:{"setConf":{"bitrate":4800,"bufferSize":8,"dataBits":"eight","parityBit":"no","stopBits":"two","cwKeyerEnable":"1",
"cwWpmSpeed":"20","rotorEnable":"","catSystemEnable":"","comPortCwType":"RTS","rtscts":"yes",
"bitrateCw":"9600","dataBitsCw":"eight","parityBitCw":"no","stopBitsCw":"one","bitrateRt":"4800",
"dataBitsRt":"eight","parityBitRt":"no","stopBitsRt":"two"},"setSerialComPath":{"value":"COM8"}}
				*/
				//////////////////////////////////////
				var dby=8;
				var sb=2;
				var br=9600;
				var pb='no';
				var rtscts=false; //flow control 				
				var com='';					
				//////////////////////////////////////
				//		ROTOR configuration 
				//////////////////////////////////////				
				if(isset(obj.setConf['rotorEnable']) && obj.setConf['rotorEnable'] == 1){
					debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
					debugc('@setConf,info,rotorEnable 1',DISPLAY_DEBUG_COM);
					if(isset(obj.setConf['bitrateRt'])){br = obj.setConf['bitrateRt'];}
					if(isset(obj.setConf['dataBitsRt'])){dby = readNumber(obj.setConf['dataBitsRt']);}
					if(isset(obj.setConf['stopBitsRt'])){sb = readNumber(obj.setConf['stopBitsRt']);}
					if(isset(obj.setConf['parityBitRt'])){pb = obj.setConf['parityBitRt'];if(obj.setConf['parityBitRt']=='no')pb='none';}
					if(isset(obj.setConf['comPortRt'])){com=obj.setConf['comPortRt'];}		
					if(isset(obj.setConf['rtsctsRt'])){
						if(obj.setConf['rtsctsRt']=='yes')rtscts = true;
						else rtscts = false;
					}							
					if(com != portNameCw){
						refresh_rt_conf_needed=1;
						debugc('@setConf,info,refresh_rt_conf_needed',DISPLAY_DEBUG_COM);
					}					
					portNameRt = com;
					var brInt = parseInt(br,10);
					setPortConfigRt	 = {
 					    path:     com,
						baudRate: brInt,
						dataBits: dby,
						stopBits: sb,
						parity: pb,
						rtscts: rtscts					
					};	
					debugc('@setConf,info,setPortConfigRt:',DISPLAY_DEBUG_COM);
					for(var x in setPortConfigRt){debugc(x+':'+setPortConfigRt[x],DISPLAY_DEBUG_COM);}
					debugc('portNameRt:'+portNameRt,DISPLAY_DEBUG_COM);
					var m1='@setConf,info,FORCE refresh ROTOR serial port configuration, and reconnect.';					
					if(connectRtId==-1 && refresh_rt_conf_needed){
						debugc(m1,DISPLAY_DEBUG_COM);
					}				
					if(connectRtId==1 && refresh_rt_conf_needed){
						debugc(m1,DISPLAY_DEBUG_COM);
						connectRtId=-1;
						if(myserialPortRt instanceof SerialPort){
							myserialPortRt.close();	
						}
					}									
					if(connectRtId != -1 && !refresh_rt_conf_needed){//already connected, configuration is the same : 
						debugc('@setConf,info, ROTOR serial port configuration is already set.',DISPLAY_DEBUG_COM);
					}					
					configuration_rt_is_ready = 1;			
				}				
				//////////////////////////////////////
				//		CW configuration 
				//////////////////////////////////////				
				if(isset(obj.setConf['cwKeyerEnable']) && obj.setConf['cwKeyerEnable'] == 1){					
					var refresh_cw_conf_needed = 0;
					debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
					debugc('@setConf,info,cwKeyerEnable 1',DISPLAY_DEBUG_COM);
					if(isset(obj.setConf['bitrateCw'])){br = obj.setConf['bitrateCw'];}
					if(isset(obj.setConf['dataBitsCw'])){dby = readNumber(obj.setConf['dataBitsCw']);}
					if(isset(obj.setConf['stopBitsCw'])){sb = readNumber(obj.setConf['stopBitsCw']);}
					if(isset(obj.setConf['parityBitCw'])){pb = obj.setConf['parityBitCw'];if(obj.setConf['parityBitCw']=='no')pb='none';}
					if(isset(obj.setConf['comPortCw'])){com=obj.setConf['comPortCw'];}		
					if(isset(obj.setConf['rtsctsCw'])){
						if(obj.setConf['rtsctsCw']=='yes')rtscts = true;
						else rtscts = false;
					}							
					if(com != portNameCw){
						refresh_cw_conf_needed=1;
						debugc('@setConf,info,refresh_cw_conf_needed',DISPLAY_DEBUG_COM);
					}					
					portNameCw = com;
					var brInt = parseInt(br,10);
					setPortConfigCw	 = {
						path:     com,
						baudRate: brInt,
						dataBits: dby,
						stopBits: sb,
						parity: pb,
						rtscts: rtscts					
					};	
					debugc('@setConf,info,setPortConfigCw:',DISPLAY_DEBUG_COM);
					for(var x in setPortConfigCw){debugc(x+':'+setPortConfigCw[x],DISPLAY_DEBUG_COM);}
					debugc('portNameCw:'+portNameCw,DISPLAY_DEBUG_COM);
					debugc('cwWpmSpeed:'+currentUserConfiguration['cwWpmSpeed'],DISPLAY_DEBUG_COM);
					debugc('comPortCwType:'+currentUserConfiguration['comPortCwType'],DISPLAY_DEBUG_COM);					
										
					var m1='@setConf,info,FORCE refresh CW serial port configuration, and reconnect.';					
					if(connectCwId==-1 && refresh_cw_conf_needed){//not connected yet : 
						debugc(m1,DISPLAY_DEBUG_COM);
					}				
					if(connectCwId==1 && refresh_cw_conf_needed){//already connected, configuration have to be update :
						debugc(m1,DISPLAY_DEBUG_COM);
						connectCwId=-1;//force refresh on next connection and close port. 
						if(myserialPortCw instanceof SerialPort){
							myserialPortCw.close();	
						}
					}									
					if(connectCwId != -1 && !refresh_cw_conf_needed){//already connected, configuration is the same : 
						debugc('@setConf,info, CW serial port configuration is already set.',DISPLAY_DEBUG_COM);
					}
					
					configuration_cw_is_ready = 1;	
					
				}

				//////////////////////////////////////
				//		CAT SYSTEM configuration 
				//////////////////////////////////////				
				if(isset(obj.setConf['catSystemEnable']) && obj.setConf['catSystemEnable'] == 1){
					debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
					debugc('@setConf,info,catSystemEnable 1',DISPLAY_DEBUG_COM);
					//set defaults:
					var refresh_cat_conf_needed=0;					
					//if client, set conf init with the client conf :
					if(isset(obj.setConf['bitrate'])){br = obj.setConf['bitrate'];}
					if(isset(obj.setConf['dataBits'])){dby = readNumber(obj.setConf['dataBits']);}
					if(isset(obj.setConf['stopBits'])){sb = readNumber(obj.setConf['stopBits']);}
					if(isset(obj.setConf['parityBit'])){pb = obj.setConf['parityBit'];if(obj.setConf['parityBit']=='no')pb='none';}
					if(isset(obj.setConf['comPort'])){com=obj.setConf['comPort'];}		
					if(isset(obj.setConf['rtscts'])){
						if(obj.setConf['rtscts']=='yes')rtscts = true;
						else rtscts = false;
					}	
					var brInt = parseInt(br,10);							
					targetRunningPortConfigCat={portNameCat:com,baudRate:brInt,dataBits:dby,stopBits:sb,parity:pb,rtscts:rtscts};				
					debugc('@setConf,info,check if update needed ... ',DISPLAY_DEBUG_COM);	
					for(var x in targetRunningPortConfigCat){
						if(x != 'portNameCat'){							
							if(x == 'stopBits' || x == 'dataBits'){
								obj.setConf[x] = readNumber(obj.setConf[x]);
							}
							if(isset(setPortConfigCat[x]) && targetRunningPortConfigCat[x] != setPortConfigCat[x]){
								debugc('@setConf,info,refresh_cat_conf_needed,runningPortConfig:'+targetRunningPortConfigCat[x],DISPLAY_DEBUG_COM);
								debugc('@setConf,info,refresh_cat_conf_needed,setConf:'+setPortConfigCat[x],DISPLAY_DEBUG_COM);
								debugc('',DISPLAY_DEBUG_COM);
								refresh_cat_conf_needed=1;
							}	
						}
					}
					if(targetRunningPortConfigCat.portNameCat != portNameCat){
							refresh_cat_conf_needed=1;
							debugc('@setConf,info,refresh_cat_conf_needed:targetRunningPortConfigCat.portNameCat:'+targetRunningPortConfigCat.portNameCat,DISPLAY_DEBUG_COM);
					}				
					var m1='@setConf,info,FORCE refresh CAT serial port configuration, and reconnect.';
					//not connected yet : 
					if(connectCatId==-1 && refresh_cat_conf_needed){
						debugc(m1,DISPLAY_DEBUG_COM);
					}				
					//already connected, configuration have to be update :
					if(connectCatId==1 && refresh_cat_conf_needed){
						debugc(m1,DISPLAY_DEBUG_COM);
						connectCatId=-1;//force refresh on next connection and close port. 
						if(myserialPortCat instanceof SerialPort){
							myserialPortCat.close();	
						}
					}				
					//already connected, configuration is the same : 
					if(connectCatId != -1 && !refresh_cat_conf_needed){
						debugc('@setConf,info, serial port configuration is already set.',DISPLAY_DEBUG_COM);
					}				
					//save configuration: 
					setPortConfigCat = {
						path: com,
						baudRate: brInt,
						dataBits: dby,
						stopBits: sb,
						parity: pb,
						rtscts: rtscts					
					};		
					if(
						targetRunningPortConfigCat.portNameCat != ''
						&& setPortConfigCat['baudRate'] != ''
					){
						portNameCat = targetRunningPortConfigCat.portNameCat;
						configuration_cat_is_ready = 1;
						debugc('@setConf,info,CAT configuration_cat_is_ready => 1',DISPLAY_DEBUG_COM);
					}			
					else{
						debugc('@setConf,info,CAT configuration_cat_is_ready => 0',DISPLAY_DEBUG_COM);
						configuration_cat_is_ready = 0;	
					}
					for(var x in setPortConfigCat){				
						debugc(x + ':' + setPortConfigCat[x],DISPLAY_DEBUG_COM);			
					}				
				}
			
				if(configuration_rt_is_ready == 1 || configuration_cw_is_ready == 1 || 	configuration_cat_is_ready == 1){
					newClient.send('ok');
				}else{
					newClient.send('ko');
				}
			
			}
			
			if("closeSerial" in obj){
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@closeSerial',DISPLAY_DEBUG_COM);
				
				if(	connectCatId	!=	-1	){
					myserialPortCat.close();
					newClient.send('ok');	
				}
				
				else{
					newClient.send('ko');	
				}
				
			}
			
			if("connect2Serial" in obj){			
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@connect2Serial',DISPLAY_DEBUG_COM);
				
				if(isset(currentUserConfiguration['catSystemEnable']) && currentUserConfiguration['catSystemEnable'] == 1){
					if(	connectCatId	==	-1	&&   configuration_cat_is_ready){				
						debugc('@connect2Serial,info,CAT port not connected ...',DISPLAY_DEBUG_COM);
						if(typeof portNameCat === 'undefined'){
							debugc('error missing portNameCat configuration.',DISPLAY_DEBUG_COM);							
							return newClient.send('ko',function(error){if(error != null) {}});
						}			
						//myserialPortCat = new SerialPort(portNameCat,setPortConfigCat);
						myserialPortCat = new SerialPort(setPortConfigCat);//change in new version 
						myserialPortCat.on('error',onErrorCatPort);
						myserialPortCat.on('open',onOpenCatPort);                    // called when the serial port opens
						myserialPortCat.on('close', function(){
							debugc('@connect2Serial,info,CAT serial port is closed now.',DISPLAY_DEBUG_COM);					
							newClient.send('ok',function(error){if(error != null) {	}});
							connectCatId=-1;
						});	
						newClient.send('ok',function(error){if(error != null) {}});
					}		
					if(	connectCatId	==	1){
						debugc('@connect2Serial,info,CAT connexion already exist,id:'+connectCatId,DISPLAY_DEBUG_COM);
						newClient.send('ok',function(error){if(error != null){}});				
					}			
					else{
						debugc('@connect2Serial,error,CAT connexion error.',DISPLAY_DEBUG_COM);
						newClient.send('ko',function(error){if(error != null){}});				
					}		
				}
				
				if(isset(currentUserConfiguration['cwKeyerEnable']) && currentUserConfiguration['cwKeyerEnable'] == 1){
					if(	connectCwId	==	-1	&&   configuration_cw_is_ready){				
						debugc('@connect2Serial,info,CW port not connected ...',DISPLAY_DEBUG_COM);
						if(typeof portNameCw === 'undefined'){
							debugc('error missing portNameCw configuration.',DISPLAY_DEBUG_COM);							
							return newClient.send('ko',function(error){if(error != null) {}});
						}			
						
						//myserialPortCw = new SerialPort(portNameCw,setPortConfigCw); old version constructor (v9)
						myserialPortCw = new SerialPort(setPortConfigCw);
						myserialPortCw.on('error',onErrorCwPort);
						myserialPortCw.on('open',onOpenCwPort);
						myserialPortCw.on('close', function(){
							debugc('@connect2Serial,info,CW serial port is closed now.',DISPLAY_DEBUG_COM);					
							newClient.send('ok',function(error){if(error != null) {	}});
							connectCwId=-1;
						});	
						newClient.send('ok',function(error){if(error != null) {}});
					}		
					if(	connectCwId	==	1){
						debugc('@connect2Serial,info,CW connexion already exist,id:'+connectCwId,DISPLAY_DEBUG_COM);
						newClient.send('ok',function(error){if(error != null){}});				
					}			
					else{
						debugc('@connect2Serial,error,CW connexion error.',DISPLAY_DEBUG_COM);
						newClient.send('ko',function(error){if(error != null){}});				
					}		
				}	
				
				if(isset(currentUserConfiguration['rotorEnable']) && currentUserConfiguration['rotorEnable'] == 1){
					if(	connectRtId	==	-1	&&   configuration_rt_is_ready){				
						debugc('@connect2Serial,info,ROTOR port not connected ...',DISPLAY_DEBUG_COM);
						if(typeof portNameRt === 'undefined'){
							debugc('error missing portNameRt configuration.',DISPLAY_DEBUG_COM);							
							return newClient.send('ko',function(error){if(error != null) {}});
						}			
						//myserialPortRt = new SerialPort(portNameRt,setPortConfigRt);//old constructor version (v9)
						myserialPortRt = new SerialPort(setPortConfigRt);
						myserialPortRt.on('error',onErrorRtPort);
						myserialPortRt.on('open',onOpenRtPort);
						myserialPortRt.on('close', function(){
							debugc('@connect2Serial,info,ROTOR serial port is closed now.',DISPLAY_DEBUG_COM);					
							newClient.send('ok',function(error){if(error != null) {	}});
							connectRtId=-1;
						});	
						newClient.send('ok',function(error){if(error != null) {}});
					}		
					if(	connectRtId	==	1){
						debugc('@connect2Serial,info,ROTOR connexion already exist,id:'+connectRtId,DISPLAY_DEBUG_COM);
						newClient.send('ok',function(error){if(error != null){}});				
					}			
					else{
						debugc('@connect2Serial,error,ROTOR connexion error.',DISPLAY_DEBUG_COM);
						newClient.send('ko',function(error){if(error != null){}});				
					}		
				}			
				
				//TODO :   send only global status : 
				//status : 
				//newClient.send('ko',function(error){if(error != null){}});				
				
			}

			if("getPortList" in obj) {
				debugc('---------------------------------------------------',DISPLAY_DEBUG_COM);
				debugc('@getSerialPortsList',DISPLAY_DEBUG_COM);	
				var reply={};
				var portList=[];
			
				SerialPort.list().then(ports => {
				   var j=0;
				   var pl='';
   				   ports.forEach(function(port) {
				   
					pl = pl + port.path + ' ';     	
					portList[j]=port.path;
					j++;     
				    	console.log(port.path);
				    	console.log(port.pnpId);
					console.log(port.manufacturer);
				        });
	
                                        c('Found port(s):'+pl);
                                  	newClient.send(JSON.stringify({setPortList:portList,connectionStatus:connectCatId}));
					
				
				  });				
		
			}

			debugc('rx client data:'+data,DISPLAY_DEBUG_NET);		

			function onOpenCatPort() {
				debugc('@onOpenCatPort() : port open',DISPLAY_DEBUG_COM);
				debugc('@onOpenCatPort() : baud rate:'+myserialPortCat.baudRate,DISPLAY_DEBUG_COM);
				connectCatId = 1;
				newClient.send('ok',function(error){if(error != null){}});
			}
			
			function onErrorCatPort(error) {
			  debugc('@onErrorCatPort,error : Serial port error'+error,DISPLAY_DEBUG_COM);			  
			  connectCatId=-1;
			  newClient.send('ko',function(error){if(error != null){}});
			}
			
			function onOpenCwPort() {
				debugc('@onOpenCwPort() : port open',DISPLAY_DEBUG_COM);
				debugc('@onOpenCwPort() : baud rate:'+myserialPortCw.baudRate,DISPLAY_DEBUG_COM);
				debugc('@onOpenCwPort() : port type:'+currentUserConfiguration['comPortCwType'],DISPLAY_DEBUG_COM);
				
				connectCwId = 1;
				
				myserialPortCw.set({dtr:false});
				myserialPortCw.set({rts:false});
				
				newClient.send('ok',function(error){if(error != null){}});
			}
			
			function onErrorCwPort(error) {
			  debugc('@onErrorCwPort,error : Serial port error'+error,DISPLAY_DEBUG_COM);			  
			  connectCwId=-1;
			  newClient.send('ko',function(error){if(error != null){}});
			}
			
			function onOpenRtPort() {
				debugc('@onOpenRtPort() : port open',DISPLAY_DEBUG_COM);
				debugc('@onOpenRtPort() : baud rate:'+myserialPortRt.baudRate,DISPLAY_DEBUG_COM);
				connectRtId = 1;
				newClient.send('ok',function(error){if(error != null){}});
			}
			
			function onErrorRtPort(error) {
			  debugc('@onErrorRtPort,error : Serial port error'+error,DISPLAY_DEBUG_COM);			  
			  connectRtId=-1;
			  newClient.send('ko',function(error){if(error != null){}});
			}			
			
		  }

	  debugc('new client request',DISPLAY_DEBUG_NET);		
	  newClient.on('message', readMessage); 
	}
	
	httpServer.listen(webSocketServerPort,()=>{});
	wss.on('connection', connectClient);
	
	c('CRX-COM Node-js server is starting, '+version+' ... ');
	c('');c(" _._     _,-'\"\"`-._");c("(,-.`._,'(       |\`-/|");
	c("    `-.-' \ )-`( , o o)");c("          `-    \`_`\"'-");c('');		
	c('Listen now on port:'+webSocketServerPort);
	c('and waiting connexion ... ');	
}
